#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/mutex.h>
#include <linux/ioport.h>
#include <linux/platform_device.h>
#include <linux/atmel_tc.h>


/*
 * This is a thin library to solve the problem of how to portably allocate
 * one of the TC blocks.  For simplicity, it doesn't currently expect to
 * share individual timers between different drivers.
 */

#if defined(CONFIG_AVR32)
/* AVR32 has these divide PBB */
const u8 atmel_tc_divisors[5] = { 0, 4, 8, 16, 32, };
EXPORT_SYMBOL(atmel_tc_divisors);

#elif defined(CONFIG_ARCH_AT91)
/* AT91 has these divide MCK */
const u8 atmel_tc_divisors[5] = { 2, 8, 32, 128, 0, };
EXPORT_SYMBOL(atmel_tc_divisors);

#endif

/* we "know" that there will be either one or two TC blocks */
static struct platform_device *blocks[2];


/**
 * atmel_tc_alloc - allocate a specified TC block
 * @block: which block to allocate
 * @iomem: used to return its IO memory resource
 *
 * Caller allocates a block.  If it is available, its I/O space is requested
 * and returned through the iomem pointer, and the device node for the block
 * is returned.  When it is not available, NULL is returned.
 *
 * On some platforms, each TC channel has its own clocks and IRQs.  Drivers
 * should clk_get() and clk_enable() "t0_clk", "t1_clk, and "t2_clk".
 * In the same vein, they should platform_get_irq() for irqs 0, 1, and 2.
 * On other platforms, only irq 0 and "t0_clk" will be available; drivers
 * should handle with both configurations.
 */
struct platform_device *atmel_tc_alloc(unsigned block, struct resource **iomem)
{
	struct platform_device	*tc;
	struct resource		*r;

	if (block >= ARRAY_SIZE(blocks) || !iomem)
		return NULL;

	tc = blocks[block];
	if (tc) {
		r = platform_get_resource(tc, IORESOURCE_MEM, 0);
		if (r)
			r = request_mem_region(r->start, 256, NULL);
		*iomem = r;
		if (!r)
			tc = NULL;
	}

	return tc;
}
EXPORT_SYMBOL_GPL(atmel_tc_alloc);

/**
 * atmel_tc_free - release a specified TC block
 * @block: which block to release
 *
 * This reverses the effect of atmel_tc_alloc(), invalidating the resource
 * returned by that routine and making the TC available to other drivers.
 */
void atmel_tc_free(struct platform_device *tc)
{
	if (tc->id >= 0 && tc->id < ARRAY_SIZE(blocks)) {
		struct resource		*r;

		r = platform_get_resource(tc, IORESOURCE_MEM, 0);
		release_mem_region(r->start, 256);
	}
}
EXPORT_SYMBOL_GPL(atmel_tc_free);

static int __init tc_probe(struct platform_device *pdev)
{
	static char __initdata e2big[] =
		KERN_ERR "tclib: can't record TC block %d\n";

	if (pdev->id < 0 || pdev->id >= ARRAY_SIZE(blocks)) {
		printk(e2big, pdev->id);
		return -ENFILE;
	}
	blocks[pdev->id] = pdev;
	return 0;
}

static struct platform_driver tc_driver = {
	.driver.name	= "atmel_tcb",
};

static int __init tc_init(void)
{
	return platform_driver_probe(&tc_driver, tc_probe);
}
arch_initcall(tc_init);
