#include <linux/module.h>
#include <linux/fs.h>
#include <linux/time.h>
#include <linux/jbd2.h>
#include <linux/highuid.h>
#include <linux/pagemap.h>
#include <linux/quotaops.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <linux/falloc.h>
#include <asm/uaccess.h>
#include <linux/ext4_jbd2.h>
#include <linux/ext4_fs_extents.h>
#include "group.h"

#define DIO_CREDITS (EXT4_RESERVE_TRANS_BLOCKS + 32)
#define EXT_SET_EXTENT_DATA(src, dest)  do {			\
		dest.block = le32_to_cpu(src->ee_block);	\
		dest.start = ext_pblock(src);			\
		dest.len = le16_to_cpu(src->ee_len);		\
					} while (0)

/*
 * this structure is used to gather extents from the tree via ioctl
 */
struct ext4_extent_buf {
	ext4_fsblk_t start;
	int buflen;
	void *buffer;
	void *cur;
	int err;
};

/*
 * this structure is used to collect stats info about the tree
 */
struct ext4_extent_tree_stats {
	int depth;
	int extents_num;
	int leaf_num;
};

int ext4_ext_walk_space(struct inode *inode, ext4_lblk_t block,
			ext4_lblk_t num, ext_prepare_callback func,
			void *cbdata)
{
	struct ext4_ext_path *path = NULL;
	struct ext4_ext_cache cbex;
	struct ext4_extent *ex;
	ext4_lblk_t next, start = 0, end = 0;
	ext4_lblk_t last = block + num;
	int depth, exists, err = 0;

	BUG_ON(func == NULL);
	BUG_ON(inode == NULL);

	while (block < last && block != EXT_MAX_BLOCK) {
		num = last - block;
		/* find extent for this block */
		path = ext4_ext_find_extent(inode, block, path);
		if (IS_ERR(path)) {
			err = PTR_ERR(path);
			path = NULL;
			break;
		}

		depth = ext_depth(inode);
		BUG_ON(path[depth].p_hdr == NULL);
		ex = path[depth].p_ext;
		next = ext4_ext_next_allocated_block(path);

		exists = 0;
		if (!ex) {
			/* there is no extent yet, so try to allocate
			 * all requested space */
			start = block;
			end = block + num;
		} else if (le32_to_cpu(ex->ee_block) > block) {
			/* need to allocate space before found extent */
			start = block;
			end = le32_to_cpu(ex->ee_block);
			if (block + num < end)
				end = block + num;
		} else if (block >= le32_to_cpu(ex->ee_block)
					+ ext4_ext_get_actual_len(ex)) {
			/* need to allocate space after found extent */
			start = block;
			end = block + num;
			if (end >= next)
				end = next;
		} else if (block >= le32_to_cpu(ex->ee_block)) {
			/*
			 * some part of requested space is covered
			 * by found extent
			 */
			start = block;
			end = le32_to_cpu(ex->ee_block)
				+ ext4_ext_get_actual_len(ex);
			if (block + num < end)
				end = block + num;
			exists = 1;
		} else {
			BUG();
		}
		BUG_ON(end <= start);

		if (!exists) {
			cbex.ec_block = start;
			cbex.ec_len = end - start;
			cbex.ec_start = 0;
			cbex.ec_type = EXT4_EXT_CACHE_GAP;
		} else {
			cbex.ec_block = le32_to_cpu(ex->ee_block);
			cbex.ec_len = ext4_ext_get_actual_len(ex);
			cbex.ec_start = ext_pblock(ex);
			cbex.ec_type = EXT4_EXT_CACHE_EXTENT;
		}

		BUG_ON(cbex.ec_len == 0);
		err = func(inode, path, &cbex, cbdata);
		ext4_ext_drop_refs(path);

		if (err < 0)
			break;
		if (err == EXT_REPEAT)
			continue;
		else if (err == EXT_BREAK) {
			err = 0;
			break;
		}

		if (ext_depth(inode) != depth) {
			/* depth was changed. we have to realloc path */
			kfree(path);
			path = NULL;
		}

		block = cbex.ec_block + cbex.ec_len;
	}

	if (path) {
		ext4_ext_drop_refs(path);
		kfree(path);
	}

	return err;
}

static int
ext4_ext_store_extent_cb(struct inode *inode,
			struct ext4_ext_path *path,
			struct ext4_ext_cache *newex,
			struct ext4_extent_buf *buf)
{

	if (newex->ec_type != EXT4_EXT_CACHE_EXTENT)
		return EXT_CONTINUE;

	if (buf->err < 0)
		return EXT_BREAK;
	if (buf->cur - buf->buffer + sizeof(*newex) > buf->buflen)
		return EXT_BREAK;

	if (!copy_to_user(buf->cur, newex, sizeof(*newex))) {
		buf->err++;
		buf->cur += sizeof(*newex);
	} else {
		buf->err = -EFAULT;
		return EXT_BREAK;
	}
	return EXT_CONTINUE;
}

static int
ext4_ext_collect_stats_cb(struct inode *inode,
			struct ext4_ext_path *path,
			struct ext4_ext_cache *ex,
			struct ext4_extent_tree_stats *buf)
{
	int depth;

	if (ex->ec_type != EXT4_EXT_CACHE_EXTENT)
		return EXT_CONTINUE;

	depth = ext_depth(inode);
	buf->extents_num++;
	if (path[depth].p_ext == EXT_FIRST_EXTENT(path[depth].p_hdr))
		buf->leaf_num++;
	return EXT_CONTINUE;
}

/**
 * ext4_ext_next_extent - search for next extent and set it to "extent"
 * @inode:	inode of the the original file
 * @path:	this will obtain data for next extent
 * @extent:	pointer to next extent we have just gotten
 *
 * This function returns 0 or 1(last_entry) if succeeded, otherwise
 * returns -EIO
 */
static int
ext4_ext_next_extent(struct inode *inode,
		     struct ext4_ext_path *path,
		     struct ext4_extent **extent)
{
	int ppos;
	int leaf_ppos = path->p_depth;

	ppos = leaf_ppos;
	if (EXT_LAST_EXTENT(path[ppos].p_hdr) > path[ppos].p_ext) {
		/* leaf block */
		*extent = ++path[ppos].p_ext;
		return 0;
	}

	while (--ppos >= 0) {
		if (EXT_LAST_INDEX(path[ppos].p_hdr) >
		    path[ppos].p_idx) {
			int cur_ppos = ppos;

			/* index block */
			path[ppos].p_idx++;
			path[ppos].p_block =
				idx_pblock(path[ppos].p_idx);
			if (path[ppos+1].p_bh)
				brelse(path[ppos+1].p_bh);
			path[ppos+1].p_bh =
				sb_bread(inode->i_sb, path[ppos].p_block);
			if (!path[ppos+1].p_bh)
				return  -EIO;
			path[ppos+1].p_hdr =
				ext_block_hdr(path[ppos+1].p_bh);

			/* halfway index block */
			while (++cur_ppos < leaf_ppos) {
				path[cur_ppos].p_idx =
					EXT_FIRST_INDEX(path[cur_ppos].p_hdr);
				path[cur_ppos].p_block =
					idx_pblock(path[cur_ppos].p_idx);
				if (path[cur_ppos+1].p_bh)
					brelse(path[cur_ppos+1].p_bh);
				path[cur_ppos+1].p_bh = sb_bread(inode->i_sb,
					path[cur_ppos].p_block);
				if (!path[cur_ppos+1].p_bh)
					return  -EIO;
				path[cur_ppos+1].p_hdr =
					ext_block_hdr(path[cur_ppos+1].p_bh);
			}

			/* leaf block */
			path[leaf_ppos].p_ext = *extent =
				EXT_FIRST_EXTENT(path[leaf_ppos].p_hdr);
			return 0;
		}
	}
	/* last_extent */
	return 1;
}

/**
 * ext4_ext_extents_info() - get extents information
 *
 * @ext_info:                   pointer to ext4_extents_info
 *  @ext_info->ino              describe an inode which is used to get extent
 *                              information
 *  @ext_info->max_entries:     defined by DEFRAG_MAX_ENT
 *  @ext_info->entries:	        amount of extents (output)
 *  @ext_info->ext[]:           array of extent (output)
 *  @ext_info->offset:          starting block offset of targeted extent
 *                              (file relative)
 *
 * @sb:                         for iget()
 *
 * This function returns 0 if next extent(s) exists,
 * or returns 1 if next extent doesn't exist, otherwise returns error value.
 */
static int ext4_ext_extents_info(struct ext4_extents_info *ext_info,
				struct super_block *sb)
{
	struct ext4_ext_path *path = NULL;
	struct ext4_extent *ext = NULL;
	struct inode *inode = NULL;
	ext4_lblk_t offset = ext_info->f_offset;
	int max_entries = ext_info->max_entries;
	int is_last_extent = 0;
	int depth = 0;
	int entries = 0;
	int err = 0;

	inode = ext4_iget(sb, ext_info->ino);
	if (IS_ERR(inode))
		return PTR_ERR(inode);

	down_write(&EXT4_I(inode)->i_data_sem);

	/* if a file doesn't exist*/
	if ((!inode->i_nlink) || (inode->i_ino < 11) ||
			!S_ISREG(inode->i_mode)) {
		ext_info->entries = 0;
		err = -ENOENT;
		goto out;
	}

	path = ext4_ext_find_extent(inode, offset, NULL);
	if (IS_ERR(path)) {
		err = PTR_ERR(path);
		path = NULL;
		goto out;
	}
	depth = ext_depth(inode);

	/* if file size is 0, skip this one. */
	if (path[depth].p_ext == NULL) {
		ext_info->entries = 0;
		goto out;
	}
	ext = path[depth].p_ext;
	EXT_SET_EXTENT_DATA(ext, ext_info->ext[entries]);
	entries = 1;

	/*
	 * The ioctl repeats this loop 'max_entries' times.
	 * So we have to call this function again if @inode had
	 * more the number of extents than 'max_entries'.
	 */
	while (entries < max_entries) {
		is_last_extent = ext4_ext_next_extent(inode, path, &ext);
		/* found next extent (not the last one)*/
		if (is_last_extent == 0) {
			EXT_SET_EXTENT_DATA(ext, ext_info->ext[entries]);
			entries++;

			/*
			 * In case @inode has > 'max_entries' extents,
			 * we must call this function again and restart from
			 * 'max_entries * n + 1'th extent.
			 * 'n' is the number of calling this function
			 * at the same @inode.
			 */
			if (entries == max_entries) {
				ext_info->f_offset =
						le32_to_cpu(ext->ee_block) +
						le32_to_cpu(ext->ee_len);
				/* check the extent is the last one or not*/
				is_last_extent =
					ext4_ext_next_extent(inode, path, &ext);
				if (is_last_extent == 1) {
					err = is_last_extent;
				} else if (is_last_extent < 0) {
					/*ERR*/
					err = is_last_extent;
					goto out;
				}
				break;
			}

		/* the extent is the last one */
		} else if (is_last_extent == 1) {
			ext_info->f_offset = 0;
			err = is_last_extent;
			break;
		} else {
			/* ERR */
			err = is_last_extent;
			goto out;
		}
	}

	ext_info->entries = entries;

out:
	if (path) {
		ext4_ext_drop_refs(path);
		kfree(path);
	}
	up_write(&EXT4_I(inode)->i_data_sem);
	iput(inode);
	return err;
}

/**
 * ext4_ext_defrag_reserve - reserve blocks for defrag
 * @inode	target inode
 * @goal	block reservation goal
 * @len		blocks count to reserve
 *
 * This function returns 0 if succeeded, otherwise
 * returns error value
 */

static int
ext4_ext_defrag_reserve(struct inode *inode, ext4_fsblk_t goal, int len)
{
	struct super_block *sb = NULL;
	handle_t *handle = NULL;
	struct buffer_head *bitmap_bh = NULL;
	struct ext4_block_alloc_info *block_i;
	struct ext4_reserve_window_node *my_rsv = NULL;
	unsigned short windowsz = 0;
	ext4_group_t group_no;
	ext4_grpblk_t grp_target_blk;
	int err = 0;

	down_write(&EXT4_I(inode)->i_data_sem);

	handle = ext4_journal_start(inode, EXT4_RESERVE_TRANS_BLOCKS);
	if (IS_ERR(handle)) {
		err = PTR_ERR(handle);
		handle = NULL;
		goto out;
	}

	if (S_ISREG(inode->i_mode) && (!EXT4_I(inode)->i_block_alloc_info)) {
		ext4_init_block_alloc_info(inode);
	} else if (!S_ISREG(inode->i_mode)) {
		printk(KERN_ERR "ext4_ext_defrag_reserve:"
				 " incorrect file type\n");
		err = -1;
		goto out;
	}

	sb = inode->i_sb;
	if (!sb) {
		printk(KERN_ERR "ext4_ext_defrag_reserve: "
				 "nonexistent device\n");
		err = -ENXIO;
		goto out;
	}
	ext4_get_group_no_and_offset(sb, goal, &group_no,
				&grp_target_blk);

	block_i = EXT4_I(inode)->i_block_alloc_info;

	if (!block_i || ((windowsz =
			block_i->rsv_window_node.rsv_goal_size) == 0)) {
		printk(KERN_ERR "ex4_ext_defrag_reserve: unable to reserve\n");
		err = -1;
		goto out;
	}

	my_rsv = &block_i->rsv_window_node;

	bitmap_bh = read_block_bitmap(sb, group_no);
	if (!bitmap_bh) {
		err = -ENOSPC;
		goto out;
	}

	BUFFER_TRACE(bitmap_bh, "get undo access for new block");
	err = ext4_journal_get_undo_access(handle, bitmap_bh);
	if (err)
		goto out;

	err = alloc_new_reservation(my_rsv, grp_target_blk, sb,
						group_no, bitmap_bh);
	if (err < 0) {
		printk(KERN_ERR "defrag: reservation faild\n");
		ext4_discard_reservation(inode);
		goto out;
	} else {
		if (len > EXT4_DEFAULT_RESERVE_BLOCKS)
			try_to_extend_reservation(my_rsv, sb,
				len - EXT4_DEFAULT_RESERVE_BLOCKS);

	}

out:
	up_write(&EXT4_I(inode)->i_data_sem);
	ext4_journal_release_buffer(handle, bitmap_bh);
	brelse(bitmap_bh);

	if (handle)
		ext4_journal_stop(handle);

	return err;
}

/**
 * ext4_ext_block_within_rsv - Is target extent reserved ?
 * @ inode       inode of target file
 * @ ex_start    start physical block number of the extent
 *               which already moved
 * @ ex_len      block length of the extent which already moved
 *
 * This function returns 0 if succeeded, otherwise
 * returns error value
 */
static int ext4_ext_block_within_rsv(struct inode *inode,
				ext4_fsblk_t ex_start, int ex_len)
{
	struct super_block *sb = inode->i_sb;
	struct ext4_block_alloc_info *block_i;
	ext4_group_t group_no;
	ext4_grpblk_t grp_blk;
	struct ext4_reserve_window_node *rsv;

	block_i = EXT4_I(inode)->i_block_alloc_info;
	if (block_i && block_i->rsv_window_node.rsv_goal_size > 0) {
		rsv = &block_i->rsv_window_node;
		if (rsv_is_empty(&rsv->rsv_window)) {
			printk(KERN_ERR "defrag: Can't defrag due to"
					" the empty reservation\n");
			return -ENOSPC;
		}
	} else {
		printk(KERN_ERR "defrag: No i_block_alloc_info\n");
		return -ENOSPC;
	}

	ext4_get_group_no_and_offset(sb, ex_start, &group_no, &grp_blk);

	if (!goal_in_my_reservation(&rsv->rsv_window, grp_blk, group_no, sb)
	    || !goal_in_my_reservation(&rsv->rsv_window, grp_blk + ex_len - 1,
		group_no, sb)){
		printk(KERN_ERR "defrag: %d or %d in bg %lu is "
				"not in rsv_window\n", grp_blk,
				grp_blk + ex_len - 1, group_no);
		return -ENOSPC;
	}
	return 0;
}

/*
 * ext4_ext_fblocks_reserve() -
 *	reserve free blocks by ext4_ext_defrag_reserve()
 * @inode:		To get a block group number
 * @ext_info:		freeblocks distribution which stored extent-like style
 *  @ext_info->ext[]    an array of struct ext4_extents_data
 */
static int ext4_ext_fblocks_reserve(struct inode *inode,
			struct ext4_extents_info *ext_info)
{
	ext4_fsblk_t ex_start = 0;
	int i;
	int ret = 0;
	int len = 0;

	for (i = 0; i < ext_info->entries; i++) {
		ex_start = ext_info->ext[i].start;
		len = ext_info->ext[i].len;

		ret = ext4_ext_defrag_reserve(inode, ex_start, len);
		if (ret < 0) {
			printk(KERN_ERR "defrag: failed "
				"ext4_ext_defrag_reserve\n");
			goto ERR;
		}
		ret = ext4_ext_block_within_rsv(inode, ex_start, len);
		if (ret < 0) {
			printk(KERN_ERR "defrag: failed "
				"ext4_ext_block_within_rsv\n");
			goto ERR;
		}
	}
	return ret;

ERR:
	down_write(&EXT4_I(inode)->i_data_sem);
	ext4_discard_reservation(inode);
	up_write(&EXT4_I(inode)->i_data_sem);
	return ret;
}

/**
 * ext4_ext_defrag_victim - Create free space for defrag
 * @filp      target file
 * @ex_info   target extents array to move
 *
 * This function returns 0 if succeeded, otherwise
 * returns error value
 */
static int ext4_ext_defrag_victim(struct file *target_filp,
			struct ext4_extents_info *ex_info)
{
	struct inode *target_inode = target_filp->f_dentry->d_inode;
	struct super_block *sb = target_inode->i_sb;
	struct file victim_file;
	struct dentry victim_dent;
	struct inode *victim_inode;
	ext4_fsblk_t goal = ex_info->goal;
	int ret = 0;
	int i = 0;
	struct ext4_extent_data ext;
	ext4_group_t group;
	ext4_grpblk_t grp_off;

	/* Setup dummy entent data */
	ext.len = 0;

	/* Get the inode of the victim file */
	victim_inode = ext4_iget(sb, ex_info->ino);
	if (IS_ERR(victim_inode))
		return PTR_ERR(victim_inode);

	/* Setup file for the victim file */
	victim_dent.d_inode = victim_inode;
	victim_file.f_dentry = &victim_dent;
	victim_file.f_mapping = victim_inode->i_mapping;

	/* Set the goal appropriate offset */
	if (goal == -1) {
		ext4_get_group_no_and_offset(victim_inode->i_sb,
				ex_info->ext[0].start, &group, &grp_off);
		goal = ext4_group_first_block_no(sb, group + 1);
	}

	for (i = 0; i < ex_info->entries; i++) {
		/* Move original blocks to another block group */
		ret = ext4_ext_defrag(&victim_file, ex_info->ext[i].block,
			ex_info->ext[i].len, goal, DEFRAG_FORCE_VICTIM, &ext);
		if (ret < 0) {
			printk(KERN_ERR "defrag: failed ext4_ext_defrag\n");
			goto ERR;
		}

		/* Sync journal blocks before reservation */
		ret = ext4_force_commit(sb);
		if (ret) {
			printk(KERN_ERR "defrag: failed ext4_force_commit (%d)\n", ret);
			goto ERR;
		}
	}

	iput(victim_inode);
	return 0;
ERR:
	down_write(&EXT4_I(target_inode)->i_data_sem);
	ext4_discard_reservation(target_inode);
	up_write(&EXT4_I(target_inode)->i_data_sem);
	iput(victim_inode);
	return ret;
}

/**
 * ext4_ext_fblocks_distribution - Search free block distribution
 * @filp      target file
 * @ex_info   ext4_extents_info
 *
 * This function returns 0 if succeeded, otherwise
 * returns error value
 */
static int ext4_ext_fblocks_distribution(struct inode *inode,
			struct ext4_extents_info *ext_info)
{
	struct buffer_head *bitmap_bh = NULL;
	struct super_block *sb = inode->i_sb;
	struct ext4_super_block *es;
	handle_t *handle;
	ext4_group_t group_no;
	ext4_grpblk_t start, end;
	ext4_fsblk_t start_block = 0;
	int num = 0;
	int len = 0;
	int i = 0;
	int err = 0;
	int block_set = 0;

	if (!sb) {
		printk(KERN_ERR "ext4_ext_fblock_distribution: "
				"nonexitent device\n");
		return -ENOSPC;
	}
	es = EXT4_SB(sb)->s_es;

	group_no = (inode->i_ino - 1) / EXT4_INODES_PER_GROUP(sb);
	start = ext_info->g_offset;
	end = EXT4_BLOCKS_PER_GROUP(sb) - 1;

	handle = ext4_journal_start(inode, 1);
	if (IS_ERR(handle)) {
		err = PTR_ERR(handle);
		return err;
	}

	bitmap_bh = read_block_bitmap(sb, group_no);
	if (!bitmap_bh) {
		err = -EIO;
		goto out;
	}

	BUFFER_TRACE(bitmap_bh, "get undo access for new block");
	err = ext4_journal_get_undo_access(handle, bitmap_bh);
	if (err)
		goto out;

	for (i = start; i <= end ; i++) {
		if (bitmap_search_next_usable_block(i, bitmap_bh, i + 1) >= 0) {
			len++;
			/* if the free block is the first one in a region */
			if (!block_set) {
				start_block =
				 i + group_no * EXT4_BLOCKS_PER_GROUP(sb);
				block_set = 1;
			}
		} else if (len) {
			ext_info->ext[num].start = start_block;
			ext_info->ext[num].len = len;
			num++;
			len = 0;
			block_set = 0;
			if (num == ext_info->max_entries) {
				ext_info->g_offset = i + 1;
				break;
			}
		}
		if ((i == end) && len) {
			ext_info->ext[num].start = start_block;
			ext_info->ext[num].len = len;
			num++;
		}
	}

	ext_info->entries = num;
out:
	ext4_journal_release_buffer(handle, bitmap_bh);
	brelse(bitmap_bh);

	if (handle)
		ext4_journal_stop(handle);

	return err;
}

int ext4_ext_ioctl(struct inode *inode, struct file *filp, unsigned int cmd,
			unsigned long arg)
{
	int err = 0;
	if (!(EXT4_I(inode)->i_flags & EXT4_EXTENTS_FL ||
		cmd == EXT4_IOC_FIBMAP))
			return -EINVAL;

	if (cmd == EXT4_IOC_GET_EXTENTS) {
		struct ext4_extent_buf buf;

		if (copy_from_user(&buf, (void *) arg, sizeof(buf)))
			return -EFAULT;

		buf.cur = buf.buffer;
		buf.err = 0;
		down_write(&EXT4_I(inode)->i_data_sem);
		err = ext4_ext_walk_space(inode, buf.start, EXT_MAX_BLOCK,
				(void *)ext4_ext_store_extent_cb, &buf);
		up_write(&EXT4_I(inode)->i_data_sem);
		if (err == 0)
			err = buf.err;
	} else if (cmd == EXT4_IOC_GET_TREE_STATS) {
		struct ext4_extent_tree_stats buf;

		down_write(&EXT4_I(inode)->i_data_sem);
		buf.depth = ext_depth(inode);
		buf.extents_num = 0;
		buf.leaf_num = 0;
		err = ext4_ext_walk_space(inode, 0, EXT_MAX_BLOCK,
				(void *)ext4_ext_collect_stats_cb, &buf);
		up_write(&EXT4_I(inode)->i_data_sem);
		if (!err)
			err = copy_to_user((void *) arg, &buf, sizeof(buf));
	} else if (cmd == EXT4_IOC_GET_TREE_DEPTH) {
		down_write(&EXT4_I(inode)->i_data_sem);
		err = ext_depth(inode);
		up_write(&EXT4_I(inode)->i_data_sem);
	} else if (cmd == EXT4_IOC_FIBMAP) {
		ext4_fsblk_t __user *p = (ext4_fsblk_t __user *)arg;
		ext4_fsblk_t block = 0;
		struct address_space *mapping = filp->f_mapping;

		if (copy_from_user(&block, (ext4_fsblk_t __user *)arg,
					sizeof(block)))
			return -EFAULT;

		lock_kernel();
		block = ext4_bmap(mapping, block);
		unlock_kernel();

		return put_user(block, p);
	} else if (cmd == EXT4_IOC_GROUP_INFO) {
		struct ext4_group_data_info grp_data;

		if (copy_from_user(&grp_data,
			(struct ext4_group_data_info __user *)arg,
			sizeof(grp_data)))
			return -EFAULT;

		grp_data.s_blocks_per_group =
			EXT4_BLOCKS_PER_GROUP(inode->i_sb);
		grp_data.s_inodes_per_group =
			EXT4_INODES_PER_GROUP(inode->i_sb);

		if (copy_to_user((struct ext4_group_data_info *)arg,
			&grp_data, sizeof(grp_data)))
			return -EFAULT;
	} else if (cmd == EXT4_IOC_FREE_BLOCKS_INFO) {
		struct ext4_extents_info ext_info;

		if (copy_from_user(&ext_info,
			(struct ext4_extents_info __user *)arg,
			sizeof(ext_info)))
			return -EFAULT;

		BUG_ON(ext_info.ino != inode->i_ino);

		err = ext4_ext_fblocks_distribution(inode, &ext_info);

		if (!err)
			err = copy_to_user((struct ext4_extents_info *)arg,
				&ext_info, sizeof(ext_info));
	} else if (cmd == EXT4_IOC_EXTENTS_INFO) {
		struct ext4_extents_info ext_info;

		if (copy_from_user(&ext_info,
				(struct ext4_extents_info __user *)arg,
				sizeof(ext_info)))
			return -EFAULT;

		err = ext4_ext_extents_info(&ext_info, inode->i_sb);
		if (err >= 0) {
			if (copy_to_user((struct ext4_extents_info __user *)arg,
				&ext_info, sizeof(ext_info)))
				return -EFAULT;
		}
	} else if (cmd == EXT4_IOC_RESERVE_BLOCK) {
		struct ext4_extents_info ext_info;

		if (copy_from_user(&ext_info,
				(struct ext4_extents_info __user *)arg,
				sizeof(ext_info)))
			return -EFAULT;

		err = ext4_ext_fblocks_reserve(inode, &ext_info);
	} else if (cmd == EXT4_IOC_MOVE_VICTIM) {
		struct ext4_extents_info ext_info;

		if (copy_from_user(&ext_info,
			(struct ext4_extents_info __user *)arg,
			sizeof(ext_info)))
			return -EFAULT;

		err = ext4_ext_defrag_victim(filp, &ext_info);

	} else if (cmd == EXT4_IOC_BLOCK_RELEASE) {
		down_write(&EXT4_I(inode)->i_data_sem);
		ext4_discard_reservation(inode);
		up_write(&EXT4_I(inode)->i_data_sem);
	} else if (cmd == EXT4_IOC_DEFRAG) {
		struct ext4_ext_defrag_data defrag;

		if (copy_from_user(&defrag,
			(struct ext4_ext_defrag_data __user *)arg,
			sizeof(defrag)))
			return -EFAULT;
		err = ext4_ext_defrag(filp, defrag.start_offset,
				defrag.defrag_size, defrag.goal, defrag.flag,
				&defrag.ext);
	}

	return err;
}

/**
 * ext4_ext_merge_across - merge extents across leaf block
 *
 * @handle	journal handle
 * @inode	target file's inode
 * @o_start	first original extent to be defraged
 * @o_end	last original extent to be defraged
 * @start_ext	first new extent to be merged
 * @new_ext	middle of new extent to be merged
 * @end_ext	last new extent to be merged
 * @flag	defrag mode (e.g. -f)
 *
 * This function returns 0 if succeed, otherwise returns error value.
 */
static int
ext4_ext_merge_across_blocks(handle_t *handle, struct inode *inode,
		struct ext4_extent *o_start,
		struct ext4_extent *o_end, struct ext4_extent *start_ext,
		struct ext4_extent *new_ext, struct ext4_extent *end_ext,
		int flag)
{
	struct ext4_ext_path *org_path = NULL;
	ext4_lblk_t eblock = 0;
	int err = 0;
	int new_flag = 0;
	int end_flag = 0;
	int defrag_flag;

	if (flag == DEFRAG_FORCE_VICTIM)
		defrag_flag = 1;
	else
		defrag_flag = 0;

	if (le16_to_cpu(start_ext->ee_len) &&
		le16_to_cpu(new_ext->ee_len) &&
		le16_to_cpu(end_ext->ee_len)) {

		if ((o_start) == (o_end)) {

			/*       start_ext   new_ext    end_ext
			 * dest |---------|-----------|--------|
			 * org  |------------------------------|
			 */

			end_flag = 1;
		} else {

			/*       start_ext   new_ext   end_ext
			 * dest |---------|----------|---------|
			 * org  |---------------|--------------|
			 */

			o_end->ee_block = end_ext->ee_block;
			o_end->ee_len = end_ext->ee_len;
			ext4_ext_store_pblock(o_end, ext_pblock(end_ext));
		}

		o_start->ee_len = start_ext->ee_len;
		new_flag = 1;

	} else if ((le16_to_cpu(start_ext->ee_len)) &&
			(le16_to_cpu(new_ext->ee_len)) &&
			(!le16_to_cpu(end_ext->ee_len)) &&
			((o_start) == (o_end))) {

		/*	 start_ext	new_ext
		 * dest |--------------|---------------|
		 * org  |------------------------------|
		 */

		o_start->ee_len = start_ext->ee_len;
		new_flag = 1;

	} else if ((!le16_to_cpu(start_ext->ee_len)) &&
			(le16_to_cpu(new_ext->ee_len)) &&
			(le16_to_cpu(end_ext->ee_len)) &&
			((o_start) == (o_end))) {

		/*	  new_ext	end_ext
		 * dest |--------------|---------------|
		 * org  |------------------------------|
		 */

		o_end->ee_block = end_ext->ee_block;
		o_end->ee_len = end_ext->ee_len;
		ext4_ext_store_pblock(o_end, ext_pblock(end_ext));

		/* If new_ext was first block */
		if (!new_ext->ee_block)
			eblock = 0;
		else
			eblock = le32_to_cpu(new_ext->ee_block);

		new_flag = 1;
	} else {
		printk(KERN_ERR "Unexpected case \n");
		return -EIO;
	}

	if (new_flag) {
		org_path = ext4_ext_find_extent(inode, eblock, NULL);
		if (IS_ERR(org_path)) {
			err = PTR_ERR(org_path);
			org_path = NULL;
			goto ERR;
		}
		err = ext4_ext_insert_extent_defrag(handle, inode,
					org_path, new_ext, defrag_flag);
		if (err)
			goto ERR;
	}

	if (end_flag) {
		org_path = ext4_ext_find_extent(inode,
				le32_to_cpu(end_ext->ee_block) - 1, org_path);
		if (IS_ERR(org_path)) {
			err = PTR_ERR(org_path);
			org_path = NULL;
			goto ERR;
		}
		err = ext4_ext_insert_extent_defrag(handle, inode,
					org_path, end_ext, defrag_flag);
		if (err)
			goto ERR;
	}
ERR:
	if (org_path) {
		ext4_ext_drop_refs(org_path);
		kfree(org_path);
	}

	return err;

}

/**
 * ext4_ext_merge_inside_block - merge new extent to the extent block
 *
 * @handle	journal handle
 * @inode	target file's inode
 * @o_start	first original extent to be defraged
 * @o_end	last original extent to be merged
 * @start_ext	first new extent to be merged
 * @new_ext	middle of new extent to be merged
 * @end_ext	last new extent to be merged
 * @eh		extent header of target leaf block
 * @replaced	the number of blocks which will be replaced with new_ext
 * @range_to_move used to dicide how to merge
 *
 * This function always returns 0.
*/
static int
ext4_ext_merge_inside_block(handle_t *handle, struct inode *inode,
		struct ext4_extent *o_start, struct ext4_extent *o_end,
		struct ext4_extent *start_ext, struct ext4_extent *new_ext,
		struct ext4_extent *end_ext, struct ext4_extent_header *eh,
		ext4_fsblk_t replaced, int range_to_move)
{
	int i = 0;
	unsigned len;

	/* Move the existing extents */
	if (range_to_move && o_end < EXT_LAST_EXTENT(eh)) {
		len = EXT_LAST_EXTENT(eh) - (o_end + 1) + 1;
		len = len * sizeof(struct ext4_extent);
		memmove(o_end + 1 + range_to_move, o_end + 1, len);
	}

	/* Insert start entry */
	if (le16_to_cpu(start_ext->ee_len))
		o_start[i++].ee_len = start_ext->ee_len;

	/* Insert new entry */
	if (le16_to_cpu(new_ext->ee_len)) {
		o_start[i].ee_block = new_ext->ee_block;
		o_start[i].ee_len = cpu_to_le16(replaced);
		ext4_ext_store_pblock(&o_start[i++], ext_pblock(new_ext));
	}

	/* Insert end entry */
	if (end_ext->ee_len)
		o_start[i] = *end_ext;

	/* Increment the total entries counter on the extent block */
	eh->eh_entries
		= cpu_to_le16(le16_to_cpu(eh->eh_entries) + range_to_move);

	return 0;
}

/**
 * ext4_ext_merge_extents - merge new extent
 *
 * @handle      journal handle
 * @inode       target file's inode
 * @org_path    path indicates first extent to be defraged
 * @o_start     first original extent to be defraged
 * @o_end       last original extent to be defraged
 * @start_ext   first new extent to be merged
 * @new_ext     middle of new extent to be merged
 * @end_ext     last new extent to be merged
 * @replaced    the number of blocks which will be replaced with new_ext
 * @flag        defrag mode (e.g. -f)
 *
 * This function returns 0 if succeed, otherwise returns error value.
 */
static int
ext4_ext_merge_extents(handle_t *handle, struct inode *inode,
		struct ext4_ext_path *org_path,
		struct ext4_extent *o_start, struct ext4_extent *o_end,
		struct ext4_extent *start_ext, struct ext4_extent *new_ext,
		struct ext4_extent *end_ext, ext4_fsblk_t replaced, int flag)
{
	struct  ext4_extent_header *eh;
	unsigned need_slots, slots_range;
	int	range_to_move, depth, ret;

	/* The extents need to be inserted
	 * start_extent + new_extent + end_extent
	 */
	need_slots = (le16_to_cpu(start_ext->ee_len) ? 1 : 0) +
			(le16_to_cpu(end_ext->ee_len) ? 1 : 0) +
			(le16_to_cpu(new_ext->ee_len) ? 1 : 0);

	/* The number of slots between start and end */
	slots_range = o_end - o_start + 1;

	/* Range to move the end of extent */
	range_to_move = need_slots - slots_range;
	depth = org_path->p_depth;
	org_path += depth;
	eh = org_path->p_hdr;

	if (depth) {
		/* Register to journal */
		ret = ext4_journal_get_write_access(handle, org_path->p_bh);
		if (ret)
			return ret;
	}

	/* expansion */
	if ((range_to_move > 0) &&
		(range_to_move > le16_to_cpu(eh->eh_max)
			- le16_to_cpu(eh->eh_entries))) {

		ret = ext4_ext_merge_across_blocks(handle, inode, o_start,
						o_end, start_ext, new_ext,
						end_ext, flag);
		if (ret < 0)
			return ret;
	} else {
		ret = ext4_ext_merge_inside_block(handle, inode, o_start,
					o_end, start_ext, new_ext, end_ext,
					eh, replaced, range_to_move);
		if (ret < 0)
			return ret;
	}

	if (depth) {
		ret = ext4_journal_dirty_metadata(handle, org_path->p_bh);
		if (ret)
			return ret;
	} else {
		ret = ext4_mark_inode_dirty(handle, inode);
		if (ret < 0)
			return ret;
	}

	return 0;

}

/**
 * ext4_ext_defrag_leaf_block -  Defragmentation for one leaf extent block.
 * @handle      journal handle
 * @org_inode   target inode
 * @org_path    path indicates first extent to be defraged
 * @dext        destination extent
 * @from        start offset on the target file
 * @flag        defrag mode (e.g. -f)
 *
 * This function returns 0 if succeed, otherwise returns error value.
 */
static int
ext4_ext_defrag_leaf_block(handle_t *handle, struct inode *org_inode,
		struct ext4_ext_path *org_path, struct ext4_extent *dext,
		ext4_lblk_t *from, int flag)
{
	unsigned long depth;
	ext4_fsblk_t replaced = 0;
	struct ext4_extent *oext, *o_start = NULL, *o_end = NULL, *prev_ext;
	struct ext4_extent new_ext, start_ext, end_ext;
	ext4_lblk_t new_end, lblock;
	unsigned short len;
	ext4_fsblk_t new_phys_end;
	int	ret;

	depth = ext_depth(org_inode);
	start_ext.ee_len = end_ext.ee_len = 0;
	o_start = o_end = oext = org_path[depth].p_ext;
	ext4_ext_store_pblock(&new_ext, ext_pblock(dext));
	new_ext.ee_len = dext->ee_len;
	len = le16_to_cpu(new_ext.ee_len);
	new_ext.ee_block = cpu_to_le32(*from);
	lblock = le32_to_cpu(oext->ee_block);
	new_end = le32_to_cpu(new_ext.ee_block)
		+ le16_to_cpu(new_ext.ee_len) - 1;
	new_phys_end = ext_pblock(&new_ext)
		+ le16_to_cpu(new_ext.ee_len) - 1;

	/* First original extent
	 * dest	 |---------------|
	 * org  |---------------|
	 */
	if (le32_to_cpu(new_ext.ee_block) >
		le32_to_cpu(oext->ee_block) &&
		le32_to_cpu(new_ext.ee_block) <
		le32_to_cpu(oext->ee_block)
		+ le16_to_cpu(oext->ee_len)) {
		start_ext.ee_len = cpu_to_le32(le32_to_cpu(new_ext.ee_block)
					- le32_to_cpu(oext->ee_block));
		replaced += le16_to_cpu(oext->ee_len)
					- le16_to_cpu(start_ext.ee_len);
	} else if (oext > EXT_FIRST_EXTENT(org_path[depth].p_hdr)) {
		/* We can merge previous extent. */
		prev_ext = oext - 1;
		if (((ext_pblock(prev_ext) + le32_to_cpu(prev_ext->ee_len))
				 == ext_pblock(&new_ext))
		 && (le32_to_cpu(prev_ext->ee_block)
			+ le32_to_cpu(prev_ext->ee_len)
				 == le32_to_cpu(new_ext.ee_block))) {
			o_start = prev_ext;
			start_ext.ee_len = cpu_to_le32(
					le16_to_cpu(prev_ext->ee_len)
					+ le16_to_cpu(new_ext.ee_len));
			new_ext.ee_len = 0;
		}
	}
		for (;;) {
		/* The extent for destination must be found. */
		BUG_ON(!oext || lblock != le32_to_cpu(oext->ee_block));
		lblock += le16_to_cpu(oext->ee_len);

		/* Middle of original extent
		 * dest |-------------------|
		 * org   |-----------------|
		 */
		if (le32_to_cpu(new_ext.ee_block) <=
			le32_to_cpu(oext->ee_block) &&
			new_end >= le32_to_cpu(oext->ee_block)
			+ le16_to_cpu(oext->ee_len) - 1)
			replaced += le16_to_cpu(oext->ee_len);

		/* Last original extent
		 * dest |----------------|
		 * org	  |---------------|
		 */
		if (new_end >= le32_to_cpu(oext->ee_block) &&
			new_end < le32_to_cpu(oext->ee_block)
				+ le16_to_cpu(oext->ee_len) - 1) {
			end_ext.ee_len
				= cpu_to_le16(le32_to_cpu(oext->ee_block)
				+ le16_to_cpu(oext->ee_len) - 1 - new_end);
			ext4_ext_store_pblock(&end_ext, (ext_pblock(o_end)
				+ cpu_to_le16(oext->ee_len)
				- cpu_to_le16(end_ext.ee_len)));
			end_ext.ee_block
				= cpu_to_le32(le32_to_cpu(o_end->ee_block)
				+ le16_to_cpu(oext->ee_len)
				- le16_to_cpu(end_ext.ee_len));
			replaced += le16_to_cpu(oext->ee_len)
				- le16_to_cpu(end_ext.ee_len);
		}

		/* Detected the block end, reached the number of replaced
		 * blocks to dext->ee_len.  Then, merge the extent.
		 */
		if (oext == EXT_LAST_EXTENT(org_path[depth].p_hdr) ||
			new_end <= le32_to_cpu(oext->ee_block)
				+ le16_to_cpu(oext->ee_len) - 1) {
			ret = ext4_ext_merge_extents(handle, org_inode,
					org_path, o_start, o_end, &start_ext,
					&new_ext, &end_ext, replaced, flag);
			if (ret < 0)
				return ret;

			/* All expected blocks are replaced */
			if (le16_to_cpu(new_ext.ee_len) <= 0) {
				if (DQUOT_ALLOC_BLOCK(org_inode, len))
					return -EDQUOT;
				return 0;
			}

			/* re-calculate new_ext */
			new_ext.ee_len = cpu_to_le32(le16_to_cpu(new_ext.ee_len)
				- replaced);
			new_ext.ee_block =
				cpu_to_le32(le32_to_cpu(new_ext.ee_block)
				+ replaced);
			ext4_ext_store_pblock(&new_ext, ext_pblock(&new_ext)
					 + replaced);
			replaced = 0;
			start_ext.ee_len = end_ext.ee_len = 0;
			o_start = NULL;

			/* All expected blocks are replaced */
			if (le16_to_cpu(new_ext.ee_len) <= 0) {
				if (DQUOT_ALLOC_BLOCK(org_inode, len))
					return -EDQUOT;
				return 0;
			}
		}

		/* Get next extent for original. */
		if (org_path)
			ext4_ext_drop_refs(org_path);
		org_path = ext4_ext_find_extent(org_inode, lblock, org_path);
		if (IS_ERR(org_path)) {
			ret = PTR_ERR(org_path);
			org_path = NULL;
			return ret;
		}
		depth = ext_depth(org_inode);
		oext = org_path[depth].p_ext;
		if (oext->ee_block + oext->ee_len <= lblock)
			return -ENOENT;

		o_end = oext;
		if (!o_start)
			o_start = oext;
	}
}

/**
 * ext4_ext_replace_branches - replace original extents with new extents.
 * @org_inode    Original inode
 * @dest_inode   temporary inode
 * @from_page    Page offset
 * @count_page   Page count to be replaced
 * @flag         defrag mode (e.g. -f)
 *
 * This function returns 0 if succeed, otherwise returns error value.
 * Replace extents for blocks from "from" to "from + count - 1".
 */
static int
ext4_ext_replace_branches(struct inode *org_inode, struct inode *dest_inode,
			pgoff_t from_page,  pgoff_t dest_from_page,
			pgoff_t count_page, int flag)
{
	handle_t *handle = NULL;
	struct ext4_ext_path *org_path = NULL;
	struct ext4_ext_path *dest_path = NULL;
	struct ext4_extent *oext, *dext, *swap_ext;
	struct ext4_extent tmp_ext, tmp_ext2;
	ext4_lblk_t from, count, dest_off, diff, org_diff;
	int err = 0;
	int depth;
	int replaced_count = 0;
	unsigned jnum;

	from = (ext4_lblk_t)from_page <<
			(PAGE_CACHE_SHIFT - dest_inode->i_blkbits);
	count = (ext4_lblk_t)count_page <<
			(PAGE_CACHE_SHIFT - dest_inode->i_blkbits);
	dest_off = (ext4_lblk_t)dest_from_page <<
			(PAGE_CACHE_SHIFT - dest_inode->i_blkbits);
	jnum = ext4_ext_writepage_trans_blocks(org_inode, count) + 3;
	handle = ext4_journal_start(org_inode, jnum);
	if (IS_ERR(handle)) {
		err = PTR_ERR(handle);
		goto out;
	}

	/* Get the original extent for the block "from" */
	org_path = ext4_ext_find_extent(org_inode, from, NULL);
	if (IS_ERR(org_path)) {
		err = PTR_ERR(org_path);
		org_path = NULL;
		goto out;
	}

	/* Get the destination extent for the head */
	dest_path = ext4_ext_find_extent(dest_inode, dest_off, NULL);
	if (IS_ERR(dest_path)) {
		err = PTR_ERR(dest_path);
		dest_path = NULL;
		goto out;
	}
	depth = ext_depth(dest_inode);
	dext = dest_path[depth].p_ext;
	/* When dext is too large, pick up the target range. */
	diff = dest_off - le32_to_cpu(dext->ee_block);
	ext4_ext_store_pblock(&tmp_ext, ext_pblock(dext) + diff);
	tmp_ext.ee_block = cpu_to_le32(le32_to_cpu(dext->ee_block) + diff);
	tmp_ext.ee_len = cpu_to_le16(le16_to_cpu(dext->ee_len) - diff);
	if (count < le16_to_cpu(tmp_ext.ee_len))
		tmp_ext.ee_len = cpu_to_le16(count);
	dext = &tmp_ext;

	depth = ext_depth(org_inode);
	oext = org_path[depth].p_ext;
	org_diff = from - le32_to_cpu(oext->ee_block);
	ext4_ext_store_pblock(&tmp_ext2, ext_pblock(oext) + org_diff);
	tmp_ext2.ee_block = tmp_ext.ee_block;

	/* adjust extent length when blocksize != pagesize */
	if (tmp_ext.ee_len <= (oext->ee_len - org_diff)) {
		tmp_ext2.ee_len = tmp_ext.ee_len;
	} else {
		tmp_ext2.ee_len = oext->ee_len - org_diff;
		tmp_ext.ee_len = tmp_ext2.ee_len;
	}
	swap_ext = &tmp_ext2;

	/* loop for the destination extents */
	while (1) {
		/* The extent for destination must be found. */
		BUG_ON(!dext || dest_off != le32_to_cpu(dext->ee_block));

		/* loop for the original extent blocks */
		err = ext4_ext_defrag_leaf_block(handle, org_inode,
						org_path, dext, &from, flag);
		if (err < 0)
			goto out;

		/* We need the function which fixes extent information for
		 * inserting.
		 * e.g. ext4_ext_merge_extents().
		 */
		err = ext4_ext_defrag_leaf_block(handle, dest_inode,
					dest_path, swap_ext, &dest_off, -1);
		if (err < 0)
			goto out;

		replaced_count += le16_to_cpu(dext->ee_len);
		dest_off += le16_to_cpu(dext->ee_len);
		from += le16_to_cpu(dext->ee_len);

		/* Already moved the expected blocks */
		if (replaced_count >= count)
			break;

		if (org_path)
			ext4_ext_drop_refs(org_path);
		org_path = ext4_ext_find_extent(org_inode, from, NULL);
		if (IS_ERR(org_path)) {
			err = PTR_ERR(org_path);
			org_path = NULL;
			goto out;
		}
		depth = ext_depth(org_inode);
		oext = org_path[depth].p_ext;
		if (oext->ee_block + oext->ee_len <= from) {
			err = 0;
			goto out;
		}

		if (dest_path)
			ext4_ext_drop_refs(dest_path);
		dest_path = ext4_ext_find_extent(dest_inode, dest_off, NULL);
		if (IS_ERR(dest_path)) {
			err = PTR_ERR(dest_path);
			dest_path = NULL;
			goto out;
		}
		depth = ext_depth(dest_inode);
		dext = dest_path[depth].p_ext;
		if (dext->ee_block + dext->ee_len <= dest_off) {
			err = 0;
			goto out;
		}

		/* When dext is too large, pick up the target range. */
		diff = dest_off - le32_to_cpu(dext->ee_block);
		ext4_ext_store_pblock(&tmp_ext, ext_pblock(dext) + diff);
		tmp_ext.ee_block =
			cpu_to_le32(le32_to_cpu(dext->ee_block) + diff);
		tmp_ext.ee_len = cpu_to_le16(le16_to_cpu(dext->ee_len) - diff);

		if ((count - replaced_count) < le16_to_cpu(tmp_ext.ee_len))
			tmp_ext.ee_len = count - replaced_count ;

		dext = &tmp_ext;

		org_diff = from - le32_to_cpu(oext->ee_block);
		ext4_ext_store_pblock(&tmp_ext2, ext_pblock(oext) + org_diff);
		tmp_ext2.ee_block = tmp_ext.ee_block;

		/* adjust extent length when blocksize != pagesize */
		if (tmp_ext.ee_len <= (oext->ee_len - org_diff)) {
			tmp_ext2.ee_len = tmp_ext.ee_len;
		} else {
			tmp_ext2.ee_len = oext->ee_len - org_diff;
			tmp_ext.ee_len = tmp_ext2.ee_len;
		}
		swap_ext = &tmp_ext2;
	}

out:
	if (handle)
		ext4_journal_stop(handle);
	if (org_path) {
		ext4_ext_drop_refs(org_path);
		kfree(org_path);
	}
	if (dest_path) {
		ext4_ext_drop_refs(dest_path);
		kfree(dest_path);
	}

	return err;
}

/**
 * ext4_ext_alloc_blocks - allocate contiguous blocks to temporary inode
 * @dest_inode   temporary inode for multiple block allocation
 * @org_inode    original inode
 * @iblock       file related offset
 * @total_blocks contiguous blocks count
 * @goal	 block offset for allocation
 * @phase        phase of create free space mode
 *
 * If succeed, fuction returns count of extent we got,
 * otherwise returns err.
 */
static int ext4_ext_alloc_blocks(struct inode *dest_inode,
		struct inode *org_inode, ext4_lblk_t iblock,
		ext4_fsblk_t total_blocks, ext4_fsblk_t goal, int phase)
{
	handle_t *handle = NULL;
	struct ext4_ext_path *dest_path = NULL;
	struct ext4_ext_path *org_path = NULL;
	struct ext4_extent newex;
	struct ext4_allocation_request ar;
	struct buffer_head *bh = NULL;
	struct super_block *org_sb = org_inode->i_sb;
	ext4_fsblk_t newblock = 0;
	ext4_fsblk_t rest = total_blocks;
	ext4_fsblk_t alloc_total = 0;
	unsigned long org_len;
	ext4_group_t dest_grp_no, org_grp_no, goal_grp_no;
	ext4_grpblk_t dest_blk_off, org_blk_off, goal_blk_off;
	int org_depth = ext_depth(org_inode);
	int metadata = 1;
	int count = 0;
	int credits = 0;
	int err = 0;
	int err2 = 0;
	int len_cnt = 0;

	ar.len = total_blocks;
	org_len = ar.len;

	/* Calculate group nubmer of org_inode block */
	if (phase == DEFRAG_FORCE_VICTIM) {
		org_path = ext4_ext_find_extent(org_inode, iblock, org_path);
		if (IS_ERR(org_path)) {
			err = PTR_ERR(org_path);
			org_path = NULL;
			goto out2;
		}
		ext4_get_group_no_and_offset(org_inode->i_sb,
				ext_pblock(org_path[org_depth].p_ext),
				&org_grp_no, &org_blk_off);
		ar.excepted_group = org_grp_no;
	} else {
		ar.excepted_group = -1;
	}

	/* Find first extent. */
	dest_path = ext4_ext_find_extent(dest_inode, iblock, dest_path);
	if (IS_ERR(dest_path)) {
		err = PTR_ERR(dest_path);
		dest_path = NULL;
		goto out2;
	}

	ar.inode = dest_inode;
	ar.flags = EXT4_MB_HINT_DATA | EXT4_MB_HINT_RESERVED
		| EXT4_MB_HINT_NOPREALLOC;
	if (goal) {
		ar.goal = goal;
	} else {
		ar.goal = ext4_ext_find_goal(dest_inode, dest_path, iblock);
	}

	ar.logical = iblock;
	ar.lleft = 0;
	ar.pleft = 0;
	ar.lright = 0;
	ar.pright = 0;

	handle = ext4_journal_start(dest_inode, credits);
	if (IS_ERR(handle)) {
		err = PTR_ERR(handle);
		goto out2;
	}

	while (alloc_total != total_blocks) {
		credits = ext4_ext_calc_credits_for_insert(dest_inode,
							dest_path);
		handle = ext4_ext_journal_restart(handle,
				credits + EXT4_TRANS_META_BLOCKS);

		if (IS_ERR(handle))
			return PTR_ERR(handle);

		newblock = ext4_mb_new_blocks(handle, &ar, &err);

		if (err) {
			/* Failed to get the contiguous blocks */
			goto out;
		} else if ((ar.len != org_len) &&
				(phase == DEFRAG_FORCE_TRY)) {
			ext4_free_blocks(handle, org_inode, newblock,
						ar.len, metadata);
			/* go to force mode */
			err = -ENOSPC;
			goto out;
		} else {
			/*
			 * If ext4_mb_new_blocks() allcates
			 * the block which used to be the metadata block,
			 * its dirty buffer_head causes the overwriting
			 * with old metadata.
			 * We should call unmap_underlying_metadata()
			 * to clear the dirty flag.
			 */
			for (len_cnt = 0; len_cnt < ar.len; len_cnt++) {
				bh = sb_find_get_block(org_sb,
						       newblock + len_cnt);
				unmap_underlying_metadata(org_sb->s_bdev,
							  newblock + len_cnt);
			}

			alloc_total += ar.len;
			ext4_get_group_no_and_offset(dest_inode->i_sb,
				goal, &goal_grp_no, &goal_blk_off);
			ext4_get_group_no_and_offset(dest_inode->i_sb,
				newblock, &dest_grp_no, &dest_blk_off);
			/* We can't allocate at the same block group */
			switch (phase) {
			case DEFRAG_FORCE_VICTIM:
				if (dest_grp_no == org_grp_no) {
					printk(KERN_ERR "defrag: Can't allocate"
					" in same block group\n");
					ext4_free_blocks(handle, org_inode,
						newblock, ar.len, metadata);
					err = -ENOSPC;
					goto out;
				}
				break;
			case DEFRAG_FORCE_GATHER:
				/* Maybe reserved blocks are already used by
				   other process */
				if (dest_grp_no != goal_grp_no
					|| alloc_total != total_blocks) {
					printk(KERN_ERR "defrag: Already used"
						" the specified blocks\n");
					ext4_free_blocks(handle, org_inode,
						newblock, ar.len, metadata);
					err = -EIO;
					goto out;
				}
				break;
			}

			newex.ee_block = cpu_to_le32(alloc_total - ar.len);
			ext4_ext_store_pblock(&newex, newblock);
			newex.ee_len = cpu_to_le16(ar.len);

			if (!phase)
				ar.goal = newblock + ar.len;
			rest = rest - ar.len;
			ar.len = rest;

			err = ext4_ext_insert_extent(handle, dest_inode,
						dest_path, &newex);
			if (!err) {
				count++;
			} else {
				ext4_free_blocks(handle, org_inode,
					newblock, ar.len, metadata);
				goto out;
			}
		}
	}

out:
	/* Faild case: We have to remove halfway blocks */
	if (err)
		err2 = ext4_ext_remove_space(dest_inode, 0);

	/* Successful case */
	if (dest_path) {
		ext4_ext_drop_refs(dest_path);
		kfree(dest_path);
	}
	if (org_path) {
		ext4_ext_drop_refs(org_path);
		kfree(org_path);
	}
out2:
	ext4_journal_stop(handle);

	if (err2) {
		return err2;
	} else if (err) {
		return err;
	}
	/* return extents count */
	return count;
}

/**
 * ext4_ext_defrag_partial - defrag original file partially
 * @filp:		pointer to file
 * @org_offset:		page index on original file
 * @dest_offset:	page index on temporary file
 * @flag:		defrag mode (e.g. -f)
 *
 * This function returns 0 if succeeded, otherwise returns error value
 */
static int
ext4_ext_defrag_partial(struct inode *tmp_inode, struct file *filp,
			pgoff_t org_offset, pgoff_t dest_offset, int flag)
{
	struct inode *inode = filp->f_dentry->d_inode;
	struct address_space *mapping = inode->i_mapping;
	struct page *page;
	pgoff_t offset_in_page = PAGE_SIZE;
	int ret = 0;

	up_write(&EXT4_I(inode)->i_data_sem);
	page = read_cache_page(inode->i_mapping, org_offset,
		(filler_t *)inode->i_mapping->a_ops->readpage, NULL);
	down_write(&EXT4_I(inode)->i_data_sem);

	if (IS_ERR(page)) {
		ret = PTR_ERR(page);
		return ret;
	}

	lock_page(page);

	/*
	 * try_to_release_page() doesn't call relasepage in writeback mode.
	 * We should care about the order of writing to the same file
	 * by multiple defrag processes.
	 * It needs to call wait_on_page_writeback() to wait for the
	 * writeback of the page.
	 */
	if (PageWriteback(page))
		wait_on_page_writeback(page);

	/* release old bh and drop refs */
	try_to_release_page(page, 0);
	ret = ext4_ext_replace_branches(inode, tmp_inode, org_offset,
			dest_offset, 1, flag);
	if (ret < 0)
		goto ERR;

	/* Clear the inode cache not to refer to the old data. */
	ext4_ext_invalidate_cache(inode);

	if (org_offset == ((inode->i_size - 1) >> PAGE_CACHE_SHIFT)) {
		offset_in_page = (inode->i_size & (PAGE_CACHE_SIZE - 1));
		/*
		 * If org_offset is the last page and i_size is
		 * multiples of PAGE_CACHE_SIZE, set PAGE_CACHE_SIZE to
		 * offset_in_page not to be 0.
		 */
		if (offset_in_page == 0)
			offset_in_page = PAGE_CACHE_SIZE;
	}

	up_write(&EXT4_I(inode)->i_data_sem);
	ret = mapping->a_ops->prepare_write(filp, page,
					0, offset_in_page);
	down_write(&EXT4_I(inode)->i_data_sem);
	if (ret)
		goto ERR;

	ret = mapping->a_ops->commit_write(filp, page,
					0, offset_in_page);
ERR:
	unlock_page(page);
	page_cache_release(page);

	return (ret < 0 ? ret : 0);
}

/**
 * ext4_ext_defrag_partial2 - defrag_partial with write_{begin, end}
 * @filp:		pointer to file
 * @org_offset:		page index on original file
 * @dest_offset:	page index on temporary file
 * @flag:		defrag mode (e.g. -f)
 *
 * This function returns 0 if succeeded, otherwise returns error value
 */
static int
ext4_ext_defrag_partial2(struct inode *tmp_inode, struct file *filp,
			pgoff_t org_offset, pgoff_t dest_offset, int flag)
{
	struct inode *inode = filp->f_dentry->d_inode;
	struct address_space *mapping = inode->i_mapping;
	struct buffer_head *bh;
	struct page *page;
	const struct address_space_operations *a_ops = mapping->a_ops;
	pgoff_t offset_in_page = PAGE_SIZE;
	int ret = 0;
	int blocksize = inode->i_sb->s_blocksize;
	int blocks_per_page = 0;
	int i = 0;
	long long offs = org_offset << PAGE_CACHE_SHIFT;
	unsigned long blk_off = 0;
	unsigned int w_flags = 0;
	void *fsdata;

	if (segment_eq(get_fs(), KERNEL_DS))
		w_flags |= AOP_FLAG_UNINTERRUPTIBLE;

	if (org_offset == ((inode->i_size - 1) >> PAGE_CACHE_SHIFT)) {
		offset_in_page = (inode->i_size & (PAGE_CACHE_SIZE - 1));
		/*
		 * If org_offset is the last page and i_size is
		 * multiples of PAGE_CACHE_SIZE, set PAGE_CACHE_SIZE to
		 * offset_in_page not to be 0.
		 */
		if (offset_in_page == 0)
			offset_in_page = PAGE_CACHE_SIZE;
	}

	up_write(&EXT4_I(inode)->i_data_sem);
	ret = a_ops->write_begin(filp, mapping, offs,
				offset_in_page, w_flags, &page, &fsdata);
	down_write(&EXT4_I(inode)->i_data_sem);

	if (unlikely(ret < 0))
		goto ERR;

	if (!PageUptodate(page)) {
		mapping->a_ops->readpage(filp, page);
		lock_page(page);
	}

	/*
	 * try_to_release_page() doesn't call relasepage in writeback mode.
	 * We should care about the order of writing to the same file
	 * by multiple defrag processes.
	 * It needs to call wait_on_page_writeback() to wait for the
	 * writeback of the page.
	 */
	if (PageWriteback(page))
		wait_on_page_writeback(page);

	/* release old bh and drop refs */
	try_to_release_page(page, 0);
	ret = ext4_ext_replace_branches(inode, tmp_inode, org_offset,
			dest_offset, 1, flag);

	if (ret < 0)
		goto ERR;

	/* Clear the inode cache not to refer to the old data. */
	ext4_ext_invalidate_cache(inode);

	if (!page_has_buffers(page))
		create_empty_buffers(page, 1 << inode->i_blkbits, 0);

	blocks_per_page = PAGE_SIZE / blocksize;
	blk_off = org_offset * blocks_per_page;

	bh = page_buffers(page);
	for (i = 0; i < blocks_per_page; i++) {
		up_write(&EXT4_I(inode)->i_data_sem);
		ret = ext4_get_block(inode, blk_off++, bh, 0);
		down_write(&EXT4_I(inode)->i_data_sem);

		if (ret < 0)
			goto ERR;

		if (bh->b_this_page != NULL)
			bh = bh->b_this_page;
	}

	ret = a_ops->write_end(filp, mapping, offs, offset_in_page,
				offset_in_page, page, fsdata);

	if (unlikely(ret < 0))
		goto ERR;
ERR:
	return (ret < 0 ? ret : 0);
}

/**
 * ext4_ext_new_extent_tree -  allocate contiguous blocks
 * @inode:		inode of the original file
 * @tmp_inode:		inode of the temporary file
 * @path:		the structure holding some info about
 *			original extent tree
 * @tar_start:		starting offset to allocate in blocks
 * @tar_blocks:		the number of blocks to allocate
 * @iblock:		file related offset
 * @goal:		block offset for allocaton
 * @flag:               phase of create free space mode
 *
 * This function returns the value as below:
 * 	0(succeeded)
 *	1(not improved)
 *	negative value(error)
 */
static int
ext4_ext_new_extent_tree(struct inode *inode, struct inode *tmp_inode,
			struct ext4_ext_path *path, ext4_lblk_t tar_start,
			ext4_lblk_t tar_blocks, ext4_lblk_t iblock,
			ext4_fsblk_t goal, int flag)
{
	struct ext4_extent *ext = NULL;
	struct ext4_extent_header *eh = NULL;
	ext4_lblk_t tar_end = tar_start + tar_blocks - 1;
	int sum_org = 0, sum_tmp = 0;
	int ret = 0, depth;
	int last_extent = 0;

	eh = ext_inode_hdr(tmp_inode);
	eh->eh_depth = 0;

	/* allocate contiguous blocks */
	sum_tmp = ext4_ext_alloc_blocks(tmp_inode, inode, iblock,
					tar_blocks, goal, flag);
	if (sum_tmp < 0) {
		ret = sum_tmp;
		goto ERR;
	}

	depth = ext_depth(inode);
	ext = path[depth].p_ext;
	while (1) {
		if (!last_extent)
			++sum_org;

		if (tar_end <= le32_to_cpu(ext->ee_block) +
			       le32_to_cpu(ext->ee_len) - 1 ||
			       last_extent) {

			if ((sum_org == sum_tmp) && !goal) {
				/* not improved */
				if (!(ret =
					ext4_ext_remove_space(tmp_inode, 0)))
					ret = 1;
			} else if (sum_org < sum_tmp &&
					flag != DEFRAG_FORCE_VICTIM) {
				/* fragment increased */
				if (!(ret =
					ext4_ext_remove_space(tmp_inode, 0)))
					ret = -ENOSPC;
				printk("defrag failed due to no space\n");
			}
			break;
		}
		if ((last_extent =
			ext4_ext_next_extent(tmp_inode,
				path, &ext)) < 0) {
			ret = last_extent;
			break;
		}
	}
ERR:
	return ret;
}

/**
 * ext4_ext_defrag - defrag whole file
 * @filp:            pointer to file
 * @from:            starting offset to defrag in blocks
 * @defrag_size:     size of defrag in blocks
 * @goal:            block offset for allocation
 * @flag:            phase of create free space mode
 * @ext:             extent to be moved (only -f)
 *
 * This function returns the number of blocks if succeeded, otherwise
 * returns error value
 */
int
ext4_ext_defrag(struct file *filp, ext4_lblk_t block_start,
		ext4_lblk_t defrag_size, ext4_fsblk_t goal,
		int flag, struct ext4_extent_data *ext)
{
	struct inode *inode = filp->f_dentry->d_inode, *tmp_inode = NULL;
	struct ext4_super_block *es = EXT4_SB(inode->i_sb)->s_es;
	struct ext4_ext_path *path = NULL, *holecheck_path = NULL;
	struct ext4_extent *ext_prev = NULL, *ext_cur = NULL, *ext_dummy = NULL;
	handle_t *handle;
	ext4_lblk_t block_end = block_start + defrag_size - 1;
	ext4_lblk_t seq_blocks = 0, seq_start = 0;
	ext4_lblk_t add_blocks = 0;
	ext4_lblk_t file_end = (inode->i_size - 1) >> inode->i_blkbits;
	pgoff_t page_offset = 0, dest_offset = 0, seq_end_page = 0;
	int ret = 0, depth = 0, last_extent = 0, seq_extents = 0;

	/* Check goal offset if goal offset was given from userspace. */
	if (((0 < goal) && (ext4_blocks_count(es) < goal)) && (goal != -1)) {
		printk(KERN_ERR "defrag: incorrect goal number %llu, "
			"you can set goal until %llu\n", goal,
			ext4_blocks_count(es));
		ret = -EINVAL;
		goto ERR1;
	}

	/* Setup for fixed blocks mode */
	if (ext->len) {
		if (ext->len < defrag_size) {
			printk("Cannot defrag due to the insufficient"
			" specified free blocks\n");
			return -EINVAL;
		}
		flag = DEFRAG_FORCE_GATHER;
		goal = ext->start;
	}

	if (file_end < block_end)
		defrag_size -= block_end - file_end;

	mutex_lock(&inode->i_mutex);
	down_write(&EXT4_I(inode)->i_data_sem);

	path = ext4_ext_find_extent(inode, block_start, NULL);
	if (IS_ERR(path)) {
		ret = PTR_ERR(path);
		path = NULL;
		goto ERR2;
	}

	/* get path structure to check hole */
	holecheck_path = ext4_ext_find_extent(inode, block_start, NULL);
	if (IS_ERR(holecheck_path)) {
		ret = PTR_ERR(holecheck_path);
		holecheck_path = NULL;
		goto ERR2;
	}

	depth = ext_depth(inode);
	ext_cur = holecheck_path[depth].p_ext;
	if (ext_cur == NULL)
		goto ERR2;

	/*
	 * if block_start was within the hole, get proper extent whose ee_block
	 * is beyond block_start
	 */
	if (le32_to_cpu(ext_cur->ee_block) +
		le32_to_cpu(ext_cur->ee_len) - 1 < block_start) {
		if ((last_extent =
				ext4_ext_next_extent(inode, holecheck_path,
				&ext_cur)) < 0) {
			ret = last_extent;
			goto ERR2;
		}
		if ((last_extent =
				ext4_ext_next_extent(inode, path,
				&ext_dummy)) < 0) {
			ret = last_extent;
			goto ERR2;
		}
	}
	seq_extents = 1;
	seq_start = ext_cur->ee_block;

	/* no blocks existed within designated range */
	if (le32_to_cpu(ext_cur->ee_block) > block_end) {
		printk("nothing done due to the lack of contiguous blocks\n");
		goto ERR2;
	}

	/* adjust start blocks */
	add_blocks = min(ext_cur->ee_block +
			 ext_cur->ee_len, block_end + 1) -
		     max(ext_cur->ee_block, block_start);

	while (!last_extent && ext_cur->ee_block <= block_end) {
		seq_blocks += add_blocks;

		handle = ext4_journal_start(inode,
			EXT4_DATA_TRANS_BLOCKS(inode->i_sb) +
			EXT4_INDEX_EXTRA_TRANS_BLOCKS + 3 +
			2 * EXT4_QUOTA_INIT_BLOCKS(inode->i_sb) + 1);
		if (IS_ERR(handle)) {
			ret = PTR_ERR(handle);
			goto ERR1;
		}
		tmp_inode = ext4_new_inode(handle,
			inode->i_sb->s_root->d_inode, S_IFREG);
		if (IS_ERR(tmp_inode)) {
			ret = -ENOMEM;
			ext4_journal_stop(handle);
			tmp_inode = NULL;
			goto ERR1;
		}

		i_size_write(tmp_inode, i_size_read(inode));
		tmp_inode->i_nlink = 0;
		ext4_ext_tree_init(handle, tmp_inode);
		ext4_orphan_add(handle, tmp_inode);
		ext4_journal_stop(handle);

		/* adjust tail blocks */
		if (seq_start + seq_blocks - 1 > block_end)
			seq_blocks = block_end - seq_start + 1;

		ext_prev = ext_cur;
		if ((last_extent =
				ext4_ext_next_extent(inode, holecheck_path,
				&ext_cur)) < 0) {
			ret = last_extent;
			break;
		}
		if (!last_extent)
			seq_extents++;
		add_blocks = le16_to_cpu(ext_cur->ee_len);

		/* found hole or reached the tail of either a designated range
		 * or the file
		 */
		if ((le32_to_cpu(ext_prev->ee_block) +
				le16_to_cpu(ext_prev->ee_len) ==
				le32_to_cpu(ext_cur->ee_block) &&
				block_end >= le32_to_cpu(ext_cur->ee_block) &&
				!last_extent)) {
			if (tmp_inode) {
				iput(tmp_inode);
				tmp_inode = NULL;
			}
			continue;
		}

		/* found an isolated block */
		if ((seq_extents == 1) && !goal) {
			seq_start = ext_cur->ee_block;
			goto CLEANUP;
		}

		ret = ext4_ext_new_extent_tree(inode, tmp_inode, path,
			seq_start, seq_blocks, block_start, goal, flag);

		if (ret < 0) {
			break;
		} else if ((ret == 1) && (!goal || (goal && !flag))) {
			ret = 0;
			seq_start = le32_to_cpu(ext_cur->ee_block);
			goto CLEANUP;
		}

		page_offset = seq_start >>
				(PAGE_CACHE_SHIFT - inode->i_blkbits);
		seq_end_page = (seq_start + seq_blocks - 1) >>
				(PAGE_CACHE_SHIFT - inode->i_blkbits);

		dest_offset = 0;
		seq_start = le32_to_cpu(ext_cur->ee_block);

		/* Discard all preallocations.
		 * This is provisional solution.
		 * When true ext4_mb_return_to_preallocation() is
		 * implemented, this will be removed.
		 */
		ext4_mb_discard_inode_preallocations(inode);

		if (inode->i_mapping->a_ops->write_begin) {
			while (page_offset <= seq_end_page) {
				/* replace original branches for new branches */
				ret = ext4_ext_defrag_partial2(tmp_inode,
							filp, page_offset,
							dest_offset, flag);
				if (ret < 0)
					goto ERR2;

				page_offset++;
				dest_offset++;
			}
		} else {
			while (page_offset <= seq_end_page) {
				/* replace original branches for new branches */
				ret = ext4_ext_defrag_partial(tmp_inode,
							filp, page_offset,
							dest_offset, flag);
				if (ret < 0)
					goto ERR2;

				page_offset++;
				dest_offset++;
			}
		}

		/* decrease buffer counter */
		if (holecheck_path)
			ext4_ext_drop_refs(holecheck_path);
		holecheck_path =
			ext4_ext_find_extent(inode, seq_start, holecheck_path);
		if (IS_ERR(holecheck_path)) {
			ret = PTR_ERR(holecheck_path);
			holecheck_path = NULL;
			break;
		}
		depth = holecheck_path->p_depth;

CLEANUP:
		/* decrease buffer counter */
		if (path)
			ext4_ext_drop_refs(path);
		path = ext4_ext_find_extent(inode, seq_start, path);
		if (IS_ERR(path)) {
			ret = PTR_ERR(path);
			path = NULL;
			break;
		}

		ext_cur = holecheck_path[depth].p_ext;
		add_blocks = le16_to_cpu(ext_cur->ee_len);
		seq_blocks = 0;
		dest_offset = 0;
		seq_extents = 1;

		if (tmp_inode) {
			iput(tmp_inode);
			tmp_inode = NULL;
		}
	}

ERR2:
	if (path) {
		ext4_ext_drop_refs(path);
		kfree(path);
	}
	if (holecheck_path) {
		ext4_ext_drop_refs(holecheck_path);
		kfree(holecheck_path);
	}
ERR1:
	up_write(&EXT4_I(inode)->i_data_sem);
	mutex_unlock(&inode->i_mutex);

	if (tmp_inode)
		iput(tmp_inode);

	return (ret ? ret : defrag_size);
}
